from __future__ import annotations
from typing import Dict, Any, List
import pandas as pd
import numpy as np


class Results:
    CLASS_NAME = "Results"

    def __init__(self, identifier: str, results_dict: Dict[str, Any]):
        self.identifier = identifier
        self.results_dict = results_dict

    def get_results(self) -> Dict[str, str]:
        return self.results_dict

    def get_key_attributes(self):
        d = self.__dict__.copy()
        d.pop("identifier", None)
        d.pop("results_dict", None)
        return d

    def get_series(self, raw):
        if raw:
            d = self.results_dict
        else:
            d = self.get_key_attributes()
        return pd.Series(d).replace(to_replace=r"^\s*$", value=np.nan, regex=True)

    def get_df(self, raw: bool = False, drop_na: bool = True) -> pd.DataFrame:
        df = pd.DataFrame({self.identifier: self.get_series(raw)})
        if drop_na:
            df.dropna(inplace=True)
        return df.T

    def __str__(self):
        d = self.get_key_attributes()
        attribute_str = "\n".join(": ".join(str(el) for el in att) for att in d.items())
        return f"{self.__class__.__name__}:\n{attribute_str}"

    def _repr_html_(self):
        return self.get_df(raw=False)._repr_html_()


class ResultsList:
    MAX_LIST = 5

    def __init__(self, results_list: List[Results]):
        self.results_list = results_list
        self.index = 0

    def __iter__(self) -> ResultsList:
        return self

    def __next__(self) -> Results:
        if self.index < len(self.results_list):
            current = self.results_list[self.index]
            self.index += 1
            return current
        else:
            self.index = 0
            raise StopIteration

    def __len__(self):
        return len(self.results_list)

    def __getitem__(self, index: int):
        return self.results_list[index]

    def __str__(self):
        vehicles_str = ",\n ".join(
            str(results) for results in self.results_list[: self.MAX_LIST]
        )
        if len(self) >= self.MAX_LIST:
            vehicles_str += ",\n..."
        return f"[{vehicles_str}]"

    def _repr_html_(self):
        return self.get_df(raw=False)._repr_html_()

    def get_results(self) -> List[Dict[str, str]]:
        return [r.get_results() for r in self.results_list]

    def get_df(self, raw: bool = False, drop_na: bool = True):
        df = pd.DataFrame()
        for rl in self.results_list:
            df[rl.identifier] = rl.get_series(raw=raw)
        if drop_na:
            df.dropna(inplace=True)
        return df.T



from typing import Dict, Any, Union
from pvaw.constants import VEHICLE_API_PATH, EARLIEST_YEAR


def get_int(num: Union[int, str]) -> int:
    try:
        return int(num)
    except ...:
        return None


def check_model_year(model_year: Union[str, int]) -> None:
    if not isinstance(model_year, (int, str)):
        raise TypeError("'model_year' must be a str or int")

    if int(model_year) < EARLIEST_YEAR:
        raise ValueError("'model_year' must be greater than 1953")


from __future__ import annotations
from typing import Dict, Union, Tuple, List
import pandas as pd
import numpy as np
from pvaw.results import Results, ResultsList
from pvaw.utils import get_int, check_model_year
from pvaw.constants import VEHICLE_API_PATH
import requests


class BatchVinDecodeError(Exception):
    pass


class Vin:
    def __init__(self, full_or_partial_vin: str, model_year: Union[str, int] = None):
        if not isinstance(full_or_partial_vin, str):
            raise TypeError('"full_or_partial_vin" must be a str')

        if len(full_or_partial_vin) > 17:
            raise ValueError('"full_or_partial_vin" must be at most 17 characters')

        self.full_or_partial_vin = full_or_partial_vin

        if model_year is not None:
            check_model_year(model_year)
            self.model_year = int(model_year)
        else:
            self.model_year = None

    def __str__(self):
        if self.model_year is not None:
            return f"{self.full_or_partial_vin},{str(self.model_year)}"
        else:
            return self.full_or_partial_vin

    def decode(self) -> Vehicle:

        args = ["format=json"]
        if self.model_year is not None:
            args.append(f"modelyear={self.model_year}")

        args_str = "&".join(args)

        path = (
            f"{VEHICLE_API_PATH}DecodeVinValues/{self.full_or_partial_vin}?{args_str}"
        )

        response = requests.get(path)
        results_dict = response.json()["Results"][0]

        return Vehicle(self, results_dict)


def decode_vins(vin_list: List[Vin]) -> ResultsList:

    if not isinstance(vin_list, list) or any(not isinstance(v, Vin) for v in vin_list):
        raise TypeError("'vin_list' must be list of Vin objects")

    if len(vin_list) == 0:
        raise ValueError("'vin_list' must have at least on Vin")

    vin_batch_str = ";".join(str(vin) for vin in vin_list)

    path = f"{VEHICLE_API_PATH}DecodeVINValuesBatch/"

    post_fields = {"format": "json", "data": vin_batch_str}

    try:
        response = requests.post(path, post_fields)
        results_list = response.json()["Results"]
    except Exception:
        raise BatchVinDecodeError("Error in API request")

    if len(results_list) != len(vin_list):
        raise BatchVinDecodeError("Incorrect number of results returned from API")

    return ResultsList(
        [Vehicle(vin_list[i], results_list[i]) for i in range(len(results_list))]
    )


class Vehicle(Results):
    def __init__(
        self,
        vin: Vin,
        results_dict: Dict[str, str],
    ):
        super().__init__(str(vin), results_dict)

        self.model_year = get_int(results_dict["ModelYear"])
        self.make = results_dict["Make"]
        self.manufacturer = results_dict["Manufacturer"]
        self.model = results_dict["Model"]
        self.full_or_partial_vin = results_dict["VIN"]
        self.vehicle_type = results_dict["VehicleType"]



from __future__ import annotations
from typing import Dict, List, Union
import pandas as pd
import requests
from pvaw.constants import VEHICLE_API_PATH
from pvaw.results import Results, ResultsList
from pvaw.utils import check_model_year


class Make(Results):
    def __init__(self, results_dict: Dict[str, str]):
        if "Make_ID" in results_dict.keys():
            self.make_id = results_dict["Make_ID"]
        else:
            self.make_id = results_dict["MakeId"]

        if "Make_Name" in results_dict.keys():
            self.make_name = results_dict["Make_Name"]
        else:
            self.make_name = results_dict["MakeName"]

        if "Mfr_Name" in results_dict.keys():
            self.manufacturer = results_dict["Mfr_Name"]
        elif "MfrName" in results_dict.keys():
            self.manufacturer = results_dict["MfrName"]
        else:
            self.manufacturer = None

        if "VehicleTypeName" in results_dict.keys():
            self.vehicle_type = results_dict["VehicleTypeName"]
        else:
            self.vehicle_type = None

        super().__init__(f"{self.make_id}-{self.manufacturer}", results_dict)


def get_makes(
    manufacturer_name_or_id: Union[str, int] = None,
    model_year: Union[str, int] = None,
    vehicle_type: str = None,
) -> ResultsList:

    if manufacturer_name_or_id is not None and not isinstance(
        manufacturer_name_or_id, (str, int)
    ):
        raise TypeError("'manufacturer_name_or_id' must be a str or int")

    if model_year is not None:
        check_model_year(model_year)

    if vehicle_type is not None and not isinstance(vehicle_type, str):
        raise TypeError("'vehicle_type' must be a str")

    if vehicle_type is not None and (manufacturer_name_or_id is not None):
        raise ValueError(
            "Cannot filter by 'vehicle_type' and 'manufacturer_name_or_id'"
        )

    if manufacturer_name_or_id is None and model_year is not None:
        raise ValueError(
            "Cannot search by 'model_year' without 'manufacturer_name_or_id'"
        )

    if manufacturer_name_or_id is not None:
        if model_year is not None:
            path = f"{VEHICLE_API_PATH}GetMakesForManufacturerAndYear/{manufacturer_name_or_id}?year={model_year}&format=json"
        else:
            path = f"{VEHICLE_API_PATH}GetMakeForManufacturer/{manufacturer_name_or_id}?format=json"
    else:
        path = f"{VEHICLE_API_PATH}GetMakesForVehicleType/{vehicle_type}?format=json"

    response = requests.get(path)
    results_list = response.json()["Results"]

    return ResultsList([Make(results_dict) for results_dict in results_list])



import requests
from typing import Dict, List
import pandas as pd
import numpy as np
from pvaw.results import Results, ResultsList
from pvaw.constants import VEHICLE_API_PATH


class WMIInfo(Results):
    def __init__(self, results_dict: Dict[str, str]) -> None:
        super().__init__(results_dict["WMI"], results_dict)
        self.wmi = results_dict["WMI"]
        self.vehicle_type = results_dict["VehicleType"]
        if "ManufacturerName" in results_dict.keys():
            self.manufacturer = results_dict["ManufacturerName"]
        else:
            self.manufacturer = results_dict["Name"]


def decode_wmi(wmi: str) -> WMIInfo:
    if not isinstance(wmi, str):
        raise TypeError("'wmi' must be a str")
    if not len(wmi) in (3, 6):
        raise ValueError(
            "'wmi' must be length 3 representing VIN position 1-3 "
            "or length 6 representing VIN positions 1-3 & 12-14."
        )

    path = f"{VEHICLE_API_PATH}DecodeWMI/{wmi}?format=json"
    response = requests.get(path)
    results_dict = response.json()["Results"][0]
    results_dict["WMI"] = wmi
    return WMIInfo(results_dict)


def get_wmis(manufacturer_search: str) -> List[WMIInfo]:
    if not isinstance(manufacturer_search, str):
        raise TypeError("'make_search' must be a str")

    path = f"{VEHICLE_API_PATH}GetWMIsForManufacturer/{manufacturer_search}?format=json"
    response = requests.get(path)
    results = response.json()["Results"]
    return ResultsList([WMIInfo(result) for result in results])



from __future__ import annotations
from typing import Dict, List, Union
import pandas as pd
import requests
from pvaw.constants import VEHICLE_API_PATH
from pvaw.results import Results, ResultsList


class Manufacturer(Results):
    def __init__(self, man_id: Union[str, int], results_dict: Dict[str, str]):
        super().__init__(man_id, results_dict)
        self.common_name = results_dict["Mfr_CommonName"]
        self.name = results_dict["Mfr_Name"]
        self.vehicle_types = [d["Name"] for d in results_dict["VehicleTypes"]]
        self.id = results_dict["Mfr_ID"]


def get_manufacturers(m_type: str = None, page: int = 1) -> ResultsList:

    args = ["format=json"]
    if m_type is not None:
        if not isinstance(m_type, str):
            raise TypeError(f"'m_type' must be a str")
        m_type_str = "%20".join(m_type.split())
        args.append(f"ManufacturerType={m_type_str}")

    if not isinstance(page, int):
        raise TypeError("'page' parameter must be an int")
    args.append(f"page={page}")

    args_str = "&".join(args)

    path = f"{VEHICLE_API_PATH}getallmanufacturers?{args_str}"

    response = requests.get(path)
    results_list = response.json()["Results"]

    return ResultsList(
        [
            Manufacturer(results_dict["Mfr_ID"], results_dict)
            for results_dict in results_list
        ]
    )


def get_manufacturer_details(manufacturer_name_or_id: Union[str, int]) -> Manufacturer:

    if not isinstance(manufacturer_name_or_id, (str, int)):
        raise TypeError("'manufacturer_name_or_id' must be a str or int")

    path = f"{VEHICLE_API_PATH}GetManufacturerDetails/{manufacturer_name_or_id}?format=json"

    response = requests.get(path)
    results_list = response.json()["Results"]

    return ResultsList(
        [
            Manufacturer(results_dict["Mfr_ID"], results_dict)
            for results_dict in results_list
        ]
    )
